from typing import Callable, Optional
from typing import Type as TypingType

from mypy.nodes import (
    MDEF,
    Block,
    ClassDef,
    NameExpr,
    PassStmt,
    SymbolTable,
    SymbolTableNode,
    TypeInfo,
    Var,
)
from mypy.plugin import ClassDefContext, Plugin
from mypy.plugins import dataclasses
from mypy.types import Instance, TypeType

MODEL_DATACLASS_TYPE_FULLNAME = "aio_rom.model.ModelDataclassType"
MODEL_FULLNAME = "aio_rom.model.Model"


def plugin(_: str) -> TypingType[Plugin]:
    return AioRomPlugin


class AioRomPlugin(Plugin):
    @staticmethod
    def transform(ctx: ClassDefContext) -> None:
        AioRomPlugin.add_exception_to_class(ctx)
        dataclasses.dataclass_class_maker_callback(ctx)

    @staticmethod
    def add_exception_to_class(ctx: ClassDefContext) -> None:
        cls: ClassDef = ctx.cls
        not_found_exception_def = ClassDef(
            "NotFoundException",
            Block([PassStmt()]),
            base_type_exprs=[NameExpr("aio_rom.exception.ModelNotFoundException")],
        )
        not_found_exception_def.fullname = f"{cls.fullname}.NotFoundException"
        not_found_class = TypeInfo(
            SymbolTable({}),
            defn=not_found_exception_def,
            module_name=cls.info.module_name,
        )
        sym = ctx.api.lookup_fully_qualified_or_none(
            "aio_rom.exception.ModelNotFoundException"
        )
        assert sym is not None
        assert isinstance(sym.node, TypeInfo)
        not_found_class.bases = [Instance(sym.node, [])]
        not_found_attr = Var(
            "NotFoundException",
            TypeType(
                Instance(sym.node, [])
            ),  # TODO: this is ModelNotFoundException, should be NotFoundException
        )
        not_found_attr.info = not_found_class
        cls.info.names["NotFoundException"] = SymbolTableNode(MDEF, not_found_attr)

    def get_base_class_hook(
        self, fullname: str
    ) -> Optional[Callable[[ClassDefContext], None]]:
        if fullname == MODEL_FULLNAME:
            return AioRomPlugin.transform
        return None

    def get_metaclass_hook(
        self, fullname: str
    ) -> Optional[Callable[[ClassDefContext], None]]:
        if fullname == MODEL_DATACLASS_TYPE_FULLNAME:
            return AioRomPlugin.transform
        return None



from contextlib import asynccontextmanager
from contextvars import ContextVar
from typing import Any, AsyncIterator, Dict, Optional

from aioredis.commands import (  # type: ignore[import]
    ContextRedis,
    Redis,
    create_redis_pool,
)
from aioredis.commands.transaction import MultiExec  # type: ignore[import]

REDIS: ContextVar[Optional[Redis]] = ContextVar("redis", default=None)
CONNECTION: ContextVar[Optional[ContextRedis]] = ContextVar("connection", default=None)
TRANSACTION: ContextVar[Optional[MultiExec]] = ContextVar("transaction", default=None)

config: Dict[str, Any] = {}


def configure(address: Optional[str] = None, *args: Any, **kwargs: Any) -> None:
    global config
    config["address"] = address
    config["args"] = args
    config["kwargs"] = kwargs


@asynccontextmanager
async def redis_pool(
    address: Optional[str] = None, *args: Any, **kwargs: Any
) -> AsyncIterator[Redis]:
    redis = REDIS.get()
    if redis:
        yield redis
        return
    redis = await create_redis_pool(
        address if address else config.get("address", "redis://localhost"),
        encoding=kwargs.get("encoding", "utf-8"),
        *args if args else config.get("args", []),
        **kwargs if kwargs else config.get("kwargs", {}),
    )
    t = REDIS.set(redis)
    try:
        yield redis
    finally:
        redis.close()
        await redis.wait_closed()
        REDIS.reset(t)


@asynccontextmanager
async def connection(
    address: Optional[str] = None, *args: Any, **kwargs: Any
) -> AsyncIterator[ContextRedis]:
    connection = CONNECTION.get()
    if connection:
        yield connection
        return

    async with redis_pool(address, *args, **kwargs) as redis:
        with await redis as conn:
            t = CONNECTION.set(conn)
            try:
                yield conn
            finally:
                CONNECTION.reset(t)


@asynccontextmanager
async def transaction() -> AsyncIterator[MultiExec]:
    transaction = TRANSACTION.get()
    if transaction:
        yield transaction
        return

    async with connection() as conn:
        transaction = conn.multi_exec()
        t = TRANSACTION.set(transaction)
        try:
            yield transaction
        finally:
            try:
                await transaction.execute()
            finally:
                TRANSACTION.reset(t)



import asyncio
import dataclasses
import logging
from dataclasses import field, fields, replace
from inspect import iscoroutinefunction, signature
from typing import (
    Any,
    AsyncIterator,
    Collection,
    Dict,
    Generic,
    List,
    Mapping,
    Tuple,
    Type,
    Union,
    cast,
)

from .exception import ModelNotFoundException
from .fields import deserialize, is_cascade, is_transient, serialize, update_field
from .session import connection, transaction
from .types import Key, M, RedisValue

_logger = logging.getLogger(__name__)


class ModelDataclassType(type, Generic[M]):
    @classmethod
    def __prepare__(  # type: ignore[override]
        mcs, name: str, bases: Tuple[type, ...], **kwds: Any
    ) -> Mapping[str, Any]:
        ns = super().__prepare__(name, bases)
        return {
            "NotFoundException": type(
                "NotFoundException", (ModelNotFoundException,), {}
            ),
            **ns,
        }

    def __new__(
        mcs,
        name: str,
        bases: Tuple[type, ...],
        namespace: Dict[str, Any],
        init: bool = True,
        repr: bool = True,
        eq: bool = True,
        order: bool = False,
        unsafe_hash: bool = False,
        frozen: bool = False,
    ) -> "ModelDataclassType[M]":
        new_namespace = namespace.copy()
        for field_name, field_type in namespace.get("__annotations__", {}).items():
            update_field(field_name, field_type, new_namespace)

        cls = cast(Type[M], super().__new__(mcs, name, bases, new_namespace))
        return dataclasses.dataclass(
            init=init,
            repr=repr,
            eq=eq,
            order=order,
            unsafe_hash=unsafe_hash,
        )(cls)


class Model(metaclass=ModelDataclassType):
    id: Key = field(init=True, repr=False, compare=False)

    @classmethod
    def prefix(cls) -> str:
        return f"{cls.__name__.lower()}"

    @classmethod
    async def get(cls: Type[M], id: Key) -> M:
        async with connection() as conn:
            db_item: Dict[str, RedisValue] = await conn.hgetall(
                f"{cls.prefix()}:{str(id)}"
            )

        if not db_item:
            raise cls.NotFoundException(f"{str(id)} not found")

        model_fields = [f for f in fields(cls) if not is_transient(f)]
        deserialized = await asyncio.gather(
            *[deserialize(f, db_item.get(f.name)) for f in model_fields]
        )

        return cls.from_dict(
            {f.name: value for f, value in zip(model_fields, deserialized)},
            strict=False,
        )

    @classmethod
    def from_dict(cls: Type[M], model: Dict[str, Any], strict: bool = True) -> M:
        parameters = signature(cls).parameters
        return (
            cls(**{k: v for k, v in model.items() if k in parameters})
            if strict
            else cls(**model)
        )

    @classmethod
    async def scan(cls: Type[M], **kwargs: Union[str, int]) -> AsyncIterator[M]:
        async with connection() as conn:
            found = set()
            async for key in conn.isscan(cls.prefix(), **kwargs):
                if key not in found:
                    value = await cls.get(key)
                    if value:
                        yield value
                        found.add(key)
                    else:
                        _logger.warning(f"{cls.__name__} Key: {key} orphaned")

    @classmethod
    async def all(cls: Type[M]) -> List[M]:
        async with connection() as conn:
            keys = await conn.smembers(cls.prefix())
            return await asyncio.gather(*[cls.get(key) for key in keys])

    @staticmethod
    async def flush() -> None:
        async with connection() as conn:
            await conn.flushdb()

    @classmethod
    async def count(cls) -> int:
        async with connection() as conn:
            return int(await conn.scard(cls.prefix()))

    @classmethod
    async def delete_all(cls: Type[M]) -> None:
        key_prefix = cls.prefix()
        async with connection() as conn:
            keys = await conn.keys(f"{key_prefix}:*")
            await conn.delete(key_prefix, *keys)

    @classmethod
    async def persisted(cls: Type[M], id: int) -> bool:
        async with connection() as conn:
            return bool(await conn.exists(f"{cls.prefix()}:{id}"))

    @property
    def db_id(self) -> str:
        return f"{self.prefix()}:{str(self.id)}"

    async def save(self, optimistic: bool = False) -> None:
        async with transaction() as tr:
            model_dict = await self._serialized_model(optimistic)
            tr.hmset_dict(self.db_id, model_dict)
            tr.sadd(self.prefix(), self.id)

    async def update(self: M, optimistic: bool = False, **changes: Any) -> M:
        async with transaction() as tr:
            model_dict = await self._serialized_model(optimistic, **changes)
            for key, value in model_dict.items():
                tr.hset(self.db_id, key, value)
            return replace(self, **changes)

    async def _serialized_model(
        self: M, optimistic: bool, **changes: Any
    ) -> Dict[str, Any]:
        async with connection() as conn:
            if optimistic:
                conn.watch(self.db_id)
            model_fields = {
                f: changes.get(f.name, getattr(self, f.name))
                for f in fields(self)
                if not is_transient(f)
                and hasattr(self, f.name)
                and (not changes or f.name in changes)
            }

            serialized = await asyncio.gather(
                *[
                    serialize(field, f"{self.db_id}:{field.name}", value)
                    for field, value in model_fields.items()
                ]
            )
            serialized_fields = list(zip(model_fields.keys(), serialized))

            await asyncio.gather(
                *[
                    value.save(optimistic=optimistic)
                    for field, value in serialized_fields
                    if iscoroutinefunction(getattr(value, "save", None))
                    and (is_cascade(field) or isinstance(value, Collection))
                ]
            )
            return {
                field.name: getattr(value, "id", f"{self.db_id}:{field.name}")
                if iscoroutinefunction(getattr(value, "save", None))
                else value
                for field, value in serialized_fields
                if value is not None
                or iscoroutinefunction(getattr(value, "save", None))
            }

    async def delete(self) -> None:
        key = self.db_id
        async with connection() as conn:
            keys = await conn.keys(f"{key}:*")
            async with transaction() as tr:
                tr.delete(*keys, key)
                tr.srem(self.prefix(), key)

    async def exists(self) -> bool:
        async with connection() as conn:
            return bool(await conn.exists(self.db_id))

    async def refresh(self: M) -> M:
        return await type(self).get(self.id)



from __future__ import annotations

import asyncio
import json
from abc import ABCMeta, abstractmethod
from dataclasses import MISSING, is_dataclass
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Collection,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    MutableSequence,
    Optional,
    Set,
    Type,
    Union,
    cast,
    overload,
)

from aioredis.commands import MultiExec  # type: ignore[import]
from typing_extensions import SupportsIndex, TypeGuard

from .exception import ModelNotLoadedException
from .session import connection, transaction
from .types import C, Key, M


def is_model(model: object) -> TypeGuard[M]:
    return is_dataclass(model) and hasattr(model, "prefix")


class RedisCollection(Collection[C], Generic[C], metaclass=ABCMeta):
    def __init__(self, key: Key, values: Collection[C], **kwargs: Any):
        self.key: Key = key
        self.values = values

    @classmethod
    async def from_key(
        cls,
        default_factory: Callable[[], Collection[C]],
        key: Key,
        eager: bool = False,
        **kwargs: Any,
    ) -> Collection[C]:
        values = await cls.get_values_for_key(key)
        if values is None:
            return default_factory() if default_factory != MISSING else None

        collection = cls(key, values, **kwargs)
        if eager:
            await collection.load()
            return collection.values
        return collection

    @abstractmethod
    async def do_save(self, tr: MultiExec, values: Collection[C]) -> None:
        pass

    async def save(self, optimistic: bool = False) -> None:
        async with connection() as conn:
            if optimistic:
                conn.watch(self.key)
            async with transaction() as tr:
                tr.delete(self.key)
                if self.values:
                    await self.do_save(tr, self.values)

    async def load(self) -> None:
        pass

    @classmethod
    @abstractmethod
    async def get_values_for_key(cls, key: Key) -> Collection[C]:
        pass

    @property
    def values(self) -> Collection[C]:
        return self._values

    @values.setter
    def values(self, values: Collection[C]) -> None:
        self._values: Collection[C] = values

    def __iter__(self) -> Iterator[C]:
        return filter(None, iter(self.values))

    def __len__(self) -> int:
        return len(self.values)

    def __contains__(self, x: object) -> bool:
        return x in self.values


class RedisSet(RedisCollection[C], Set[C], Generic[C]):
    @classmethod
    async def get_values_for_key(cls, key: Key) -> Set[C]:
        async with connection() as conn:
            return {json.loads(value) for value in await conn.smembers(key)}

    async def do_save(self, tr: MultiExec, values: Collection[C]) -> None:
        tr.sadd(self.key, *values)

    def add(self, value: C) -> None:
        cast(Set[C], super().values).add(value)

    def discard(self, value: C) -> None:
        cast(Set[C], super().values).discard(value)


class RedisList(RedisCollection[C], MutableSequence[C], Generic[C]):
    @classmethod
    async def get_values_for_key(cls, key: Key) -> List[C]:
        async with connection() as redis:
            return [json.loads(value) for value in await redis.lrange(key, 0, -1)]

    async def do_save(self, tr: MultiExec, values: Collection[C]) -> None:
        tr.rpush(self.key, *values)

    def insert(self, index: int, value: C) -> None:
        cast(List[C], super().values).insert(index, value)

    @overload
    def __getitem__(self, i: int) -> C:
        ...

    @overload
    def __getitem__(self, i: slice) -> List[C]:
        ...

    def __getitem__(self, i: Union[int, slice]) -> Union[C, List[C]]:
        return cast(List[C], super().values).__getitem__(i)

    @overload
    def __setitem__(self, index: int, o: C) -> None:
        ...

    @overload
    def __setitem__(self, index: slice, o: Iterable[C]) -> None:
        ...

    def __setitem__(self, i: Union[int, slice], o: Union[C, Iterable[C]]) -> None:
        if isinstance(i, slice) and isinstance(o, Iterable):
            cast(List[C], super().values).__setitem__(i, o)
        elif isinstance(i, int) and not isinstance(o, Iterable):
            cast(List[C], super().values).insert(i, o)

    def __delitem__(self, i: Union[int, slice]) -> None:
        cast(List[C], super().values).__delitem__(i)


class ModelCollection(RedisCollection[M], AsyncIterator[M], metaclass=ABCMeta):
    def __init__(
        self,
        key: Key,
        keys: Collection[Union[Key, M]],
        model_class: Type[M],
        cascade: bool,
    ):
        super().__init__(
            key, cast(List[M], [value for value in keys if is_model(value)])
        )
        self._cache.update(
            {key: None for key in keys if isinstance(key, (int, str, bytes))}
        )
        self._model_class = model_class
        self._cascade = cascade

    @classmethod
    def serialize(
        cls,
        key: Key,
        keys: Union[Collection[Key], ModelCollection[M]],
        model_class: Type[M],
        cascade: bool = False,
    ) -> ModelCollection[M]:
        return (
            keys
            if isinstance(keys, ModelCollection)
            else cls(key, keys, model_class, cascade)
        )

    async def do_save(self, tr: MultiExec, values: Collection[M]) -> None:
        await super().do_save(tr, list(self._cache.keys()))  # type: ignore[arg-type]

    async def save(self, optimistic: bool = False) -> None:
        async with transaction():
            await super().save(optimistic=optimistic)
            if self._cascade and self.values:
                await asyncio.gather(
                    *[v.save(optimistic=optimistic) for v in self.values]
                )

    async def load(self) -> None:
        async with connection():
            await asyncio.gather(*[self._get_item(key) for key in self._cache.keys()])

    @property
    def values(self) -> Collection[M]:
        return list(filter(None, self._cache.values()))

    @values.setter
    def values(self, values: Collection[M]) -> None:
        self._cache: Dict[Key, Optional[M]] = {v.id: v for v in values}

    async def _get_item(self, key: Key) -> M:
        value = self._cache.get(key, None)
        if not value:
            value = await self._model_class.get(key)
            self._cache[key] = value
        return value

    def __aiter__(self) -> AsyncIterator[M]:
        self._iter = map(self._get_item, self._cache.keys())
        return self

    async def __anext__(self) -> M:
        try:
            return await next(self._iter)
        except StopIteration:
            raise StopAsyncIteration


class RedisModelSet(ModelCollection[M], RedisSet[M], Generic[M]):
    @property
    def values(self) -> Set[M]:
        return set(super().values)

    @values.setter
    def values(self, values: Collection[M]) -> None:
        super(RedisModelSet, type(self)).values.fset(self, values)  # type: ignore[attr-defined] # noqa: E501

    def add(self, value: M) -> None:
        self._cache[value.id] = value

    def discard(self, value: M) -> None:
        self._cache.pop(value.id)


class RedisModelList(ModelCollection[M], RedisList[M], Generic[M]):
    @property
    def values(self) -> List[M]:
        return list(super().values)

    @values.setter
    def values(self, values: Collection[C]) -> None:
        super(RedisModelList, type(self)).values.fset(self, values)  # type: ignore[attr-defined] # noqa: E501

    def _get_cached_item(self, key: Key) -> M:
        if key in self._cache:
            value = self._cache[key]
            if not value:
                raise ModelNotLoadedException(
                    f"{self._model_class.__name__} id {key!r} was not loaded"
                )
            return value
        else:
            raise KeyError

    def insert(self, index: int, value: M) -> None:
        super().insert(index, value)
        self._cache[value.id] = value

    @overload
    def __getitem__(self, index: int) -> M:
        ...

    @overload
    def __getitem__(self, index: slice) -> List[M]:
        ...

    def __getitem__(self, i: Union[int, slice]) -> Union[M, List[M]]:
        if isinstance(i, slice):
            return [
                self._get_cached_item(super().__getitem__(index).id)
                for index in range(*i.indices(len(self)))
            ]
        else:
            return self._get_cached_item(super().__getitem__(i).id)

    @overload
    def __setitem__(self, index: int, o: M) -> None:
        ...

    @overload
    def __setitem__(self, index: slice, o: Iterable[M]) -> None:
        ...

    def __setitem__(self, i: Union[int, slice], o: Union[M, Iterable[M]]) -> None:
        if isinstance(i, slice) and isinstance(o, SupportsIndex):
            for obj_index, index in enumerate(range(*i.indices(len(self)))):
                self.insert(index, o[obj_index])
        elif not isinstance(i, slice) and not isinstance(o, Iterable):
            self.insert(i, o)
        else:
            raise ValueError

    def __delitem__(self, i: Union[int, slice]) -> None:
        if isinstance(i, slice):
            for index in range(*i.indices(len(self))):
                del self._cache[super().__getitem__(index).id]
        else:
            del self._cache[super().__getitem__(i).id]
        super().__delitem__(i)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aio_rom']

package_data = \
{'': ['*']}

install_requires = \
['aioredis==1.3.1', 'typing-extensions>=3.10.0.0,<4.0.0.0']

extras_require = \
{'mypy': ['mypy>=0.910,<0.911']}

setup_kwargs = {
    'name': 'aio-rom',
    'version': '0.1.1',
    'description': 'asyncio based Redis object mapper',
    'long_description': 'Python Redis Object Mapper\n======================\n\nasyncio based Redis object mapper\n\n## Table of content\n\n- [Installation](#installation)\n- [Usage](#usage)\n- [Features](#usage)\n- [TODO](#todo)\n- [Limitations](#limitations)\n\n## Installation\n\nTODO\n\n## Usage\n\n```python\nimport asyncio\n\nfrom dataclasses import field\nfrom typing import Set, Dict\n\nfrom aio_rom import Model\nfrom aio_rom.fields import Metadata\nfrom aio_rom.session import redis_pool\n\n\nclass Foo(Model):\n    bar: int\n    foobar: Set[int] = field(default_factory=set)\n    my_boolean: bool = False\n    transient_field: Dict = field(metadata=Metadata(transient=True))\n\n\nclass OtherFoo(Model):\n    foo: Foo\n\nasync def main():\n    async with redis_pool("redis://localhost"):\n        foo = Foo(123, {1,2,3}, True)\n        await foo.save()\n        ...\n        foo2 = await Foo.get(321)\n        other_foo = OtherFoo(303, foo2)\n        await other_foo.save()\n\nasyncio.run(main())\n```\n## Features\nTODO\n\n## TODO\n1. Docs\n1. Tests\n\n## Limitations\n1. `configure` must be called before other calls to Redis can succeed, no defaults to localhost atm.\n1. You cannot use `from __future__ import annotations` in the same file you define your models. See https://bugs.python.org/issue39442\n1. TODO Supported datatypes\n1. Probably more ...\n',
    'author': 'Federico Jaite',
    'author_email': 'fede_654_87@hotmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/fedej/aio-rom',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import json
from asyncio.coroutines import iscoroutine
from collections.abc import MutableSequence
from dataclasses import Field
from dataclasses import field as dc_field
from functools import partial
from types import MappingProxyType
from typing import (
    AbstractSet,
    Any,
    Awaitable,
    Callable,
    Dict,
    Optional,
    Type,
    Union,
    cast,
)

from typing_extensions import TypedDict, TypeGuard, get_args, get_origin

from .attributes import RedisList, RedisModelList, RedisModelSet, RedisSet, is_model
from .types import (
    Deserialized,
    Deserializer,
    F,
    Key,
    M,
    RedisValue,
    Serializable,
    Serializer,
)


class Metadata(TypedDict, total=False):
    transient: bool
    cascade: bool
    eager: bool
    optional: bool
    serializer: Serializer[Any]
    deserializer: Deserializer[Any]


def is_transient(dataclass_field: Field[F]) -> bool:
    return dataclass_field.metadata.get("transient", False)


def is_eager(dataclass_field: Field[F]) -> bool:
    return dataclass_field.metadata.get("eager", False)


def is_cascade(dataclass_field: Field[F]) -> bool:
    return dataclass_field.metadata.get("cascade", False)


def is_optional(dataclass_field: Field[F]) -> bool:
    return dataclass_field.metadata.get("optional", False)


def is_model_type(model_type: Type[Any]) -> TypeGuard[Type[M]]:
    return is_model(model_type) and isinstance(model_type, type)


def _deserialize_reference(
    model_class: Type[M],
) -> Callable[[Key], Awaitable[Optional[M]]]:
    async def deserializer(
        key: Key,
    ) -> Optional[M]:
        if key is None:
            return None
        return await model_class.get(key)

    return deserializer


async def deserialize(
    dataclass_field: Field[F], value: Optional[RedisValue]
) -> Optional[F]:
    if value is not None:
        metadata = cast(Metadata, dataclass_field.metadata)
        deserializer = metadata["deserializer"]
        deserialized_value: Deserialized[F] = deserializer(value)
        if iscoroutine(deserialized_value):
            return await cast(Awaitable[F], deserialized_value)
        else:
            return cast(F, deserialized_value)
    elif is_optional(dataclass_field):
        return None
    else:
        raise TypeError(f"Value missing for required field {dataclass_field.name}")


def _is_coroutine_serializable(val: Any) -> TypeGuard[Awaitable[Serializable]]:
    return iscoroutine(val)


async def serialize(dataclass_field: Field[F], key: str, value: F) -> Serializable:
    metadata = cast(Metadata, dataclass_field.metadata)
    serializer = metadata["serializer"]
    val = serializer(key, value)
    return await val if _is_coroutine_serializable(val) else cast(Serializable, val)


def pass_through_serializer(_: str, value: F) -> F:
    return value


def pass_through_deserializer(value: Any) -> Any:
    return value


def update_field(
    name: str, field_type: Type[Union[F, M]], fields: Dict[str, Field[F]]
) -> None:
    dataclass_field = fields.get(name, dc_field()) or dc_field()
    eager = is_eager(dataclass_field)
    cascade = is_cascade(dataclass_field)

    optional = False
    if type(None) in get_args(field_type):
        # Unwrap optional type
        optional = True
        field_type = get_args(field_type)[0]

    origin = get_origin(field_type) or field_type
    type_args = get_args(field_type)

    deserializer: Union[Deserializer[M], Deserializer[F]] = json.loads
    serializer: Serializer[F] = lambda key, value: json.dumps(value)

    if isinstance(field_type, type) and issubclass(field_type, str):
        deserializer = pass_through_deserializer
        serializer = pass_through_serializer
    elif is_model_type(field_type):
        deserializer = cast(Deserializer[M], _deserialize_reference(field_type))
        serializer = pass_through_serializer

    if isinstance(origin, type):
        if issubclass(origin, AbstractSet):
            model_class = type_args[0]
            if is_model_type(model_class):
                deserializer = cast(
                    Callable[[], F],
                    partial(
                        RedisModelSet.from_key,
                        dataclass_field.default_factory,  # type: ignore[misc]
                        model_class=model_class,
                        eager=eager,
                        cascade=cascade,
                    ),
                )
                serializer = partial(
                    RedisModelSet.serialize, model_class=model_class, cascade=cascade
                )
            else:
                deserializer = cast(
                    Callable[[], F],
                    partial(
                        RedisSet.from_key,
                        dataclass_field.default_factory,  # type: ignore[misc]
                        eager=eager,
                    ),
                )
                serializer = RedisSet
        elif issubclass(origin, MutableSequence):
            model_class = type_args[0]
            if is_model(model_class):
                deserializer = cast(
                    Callable[[], F],
                    partial(
                        RedisModelList.from_key,
                        dataclass_field.default_factory,  # type: ignore[misc]
                        model_class=model_class,
                        eager=eager,
                        cascade=cascade,
                    ),
                )
                serializer = partial(
                    RedisModelList.serialize, model_class=type_args[0], cascade=cascade
                )
            else:
                deserializer = cast(
                    Callable[[], F],
                    partial(
                        RedisList.from_key,
                        dataclass_field.default_factory,  # type: ignore[misc]
                        eager=eager,
                    ),
                )
                serializer = RedisList

    metadata = cast(Metadata, dict(dataclass_field.metadata)).copy()
    metadata.update(
        {
            "deserializer": deserializer,
            "serializer": serializer,
            "optional": optional,
        }
    )
    dataclass_field.metadata = MappingProxyType(metadata)
    fields[name] = dataclass_field



from typing import TYPE_CHECKING, Awaitable, Callable, Optional, TypeVar, Union

if TYPE_CHECKING:
    from .attributes import RedisCollection
    from .model import Model

F = TypeVar("F", str, bool, int, float)
M = TypeVar("M", bound="Model")
C = TypeVar("C", bound=Union[str, bool, int, float, "Model"])

RedisValue = Union[str, bytes]
Key = Union[int, RedisValue]
Serializable = Union[RedisValue, "Model", "RedisCollection"]
Serializer = Callable[..., Union[F, Serializable, Awaitable[Serializable]]]
Deserialized = Union[Optional[F], Awaitable[Optional[F]]]
Deserializer = Callable[..., Deserialized[F]]
